//Pyjama compiler version:v2.3.0
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import pj.*;

import pj.pr.*;
import pj.PjRuntime;
import pj.Pyjama;
import pi.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;
import java.lang.reflect.InvocationTargetException;
import pj.pr.exceptions.*;

public class PjMandelbrotPar extends JFrame implements MouseListener, MouseWheelListener, KeyListener {

    static {
        Pyjama.omp_create_virtual_target("worker", 1);
        Pyjama.omp_set_num_threads(2);
    }

    private static final long serialVersionUID = -7311452299949615843L;

    private static final int iterationDept = 2550;

    private BufferedImage display = null;

    private WritableRaster raster = null;

    private int frameWidth = 1360;

    private int frameHight = 720;

    private int mouseX = (int) frameWidth / 2;

    private int mouseY = (int) frameHight / 2;

    private double xmin = MandelbrotDefault.xmin;

    private double xmax = MandelbrotDefault.xmax;

    private double ymin = MandelbrotDefault.ymin;

    private double ymax = MandelbrotDefault.ymax;

    private final int zoomScale = 10;

    enum ZoomType {

        zoomIn, zoomOut, noZoom
    }

    ;

    class MandelbrotDefault {

        static final double xmin = -2.5;

        static final double xmax = 1.0;

        static final double ymin = -1.25;

        static final double ymax = 1.25;
    }

    public PjMandelbrotPar() {
        super("Mandelbrot");
        setSize(this.frameWidth, this.frameHight);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setResizable(true);
        final Component graphics = new Component() {

            private static final long serialVersionUID = 1L;

            public void paint(Graphics g) {{
                g.drawImage(display, 0, 0, null);
            }
            }

        };
        display = new BufferedImage(this.frameWidth, this.frameHight, BufferedImage.TYPE_INT_RGB);
        raster = display.getRaster();
        graphics.addMouseListener(this);
        graphics.addMouseWheelListener(this);
        this.addKeyListener(this);
        graphics.addComponentListener(new ComponentAdapter() {

            public void componentResized(ComponentEvent e) {{
                frameWidth = graphics.getWidth();
                frameHight = graphics.getHeight();
                updatePosition(ZoomType.noZoom);
            }
            }

        });
        getContentPane().add(graphics);
        updateRaster(this.xmax, this.ymax, this.xmin, this.ymin, this.frameWidth, this.frameHight);
        this.repaint();
        this.setVisible(true);
    }

    private void updatePosition(ZoomType zoomType) {{
        double xScale = (this.xmax - this.xmin) / this.frameWidth;
        double yScale = (this.ymax - this.ymin) / this.frameHight;
        double centreX = this.xmin + (this.xmax - this.xmin) * this.mouseX / this.frameWidth;
        double centreY = this.ymin + (this.ymax - this.ymin) * this.mouseY / this.frameHight;
        double scale = zoomScale;
        if (zoomType == ZoomType.zoomOut) {
            scale = 1 / scale;
        } else if (zoomType == ZoomType.noZoom) {
            scale = 1;
        }
        this.xmax = centreX + this.frameWidth / 2 * xScale / scale;
        this.xmin = centreX - this.frameWidth / 2 * xScale / scale;
        this.ymax = centreY + this.frameHight / 2 * yScale / scale;
        this.ymin = centreY - this.frameHight / 2 * yScale / scale;
        updateRaster(this.xmax, this.ymax, this.xmin, this.ymin, this.frameWidth, this.frameHight);
    }
    }


    private void updateRaster(double xmax, double ymax, double xmin, double ymin, int width, int hight) {{
        double xTick = (xmax - xmin) / width;
        double yTick = (ymax - ymin) / hight;
        double x = xmin;
        double y = ymin;
        /*OpenMP Parallel region (#0) -- START */
        InternalControlVariables icv_previous__OMP_ParallelRegion_0 = PjRuntime.getCurrentThreadICV();
        InternalControlVariables icv__OMP_ParallelRegion_0 = PjRuntime.inheritICV(icv_previous__OMP_ParallelRegion_0);
        int _threadNum__OMP_ParallelRegion_0 = icv__OMP_ParallelRegion_0.nthreads_var.get(icv__OMP_ParallelRegion_0.levels_var);
        _OMP_ParallelRegion_0 _OMP_ParallelRegion_0_in = new _OMP_ParallelRegion_0(_threadNum__OMP_ParallelRegion_0,icv__OMP_ParallelRegion_0);
        _OMP_ParallelRegion_0_in.width = width;
        _OMP_ParallelRegion_0_in.xTick = xTick;
        _OMP_ParallelRegion_0_in.x = x;
        _OMP_ParallelRegion_0_in.hight = hight;
        _OMP_ParallelRegion_0_in.yTick = yTick;
        _OMP_ParallelRegion_0_in.y = y;
        _OMP_ParallelRegion_0_in.runParallelCode();
        width = _OMP_ParallelRegion_0_in.width;
        xTick = _OMP_ParallelRegion_0_in.xTick;
        x = _OMP_ParallelRegion_0_in.x;
        hight = _OMP_ParallelRegion_0_in.hight;
        yTick = _OMP_ParallelRegion_0_in.yTick;
        y = _OMP_ParallelRegion_0_in.y;
        PjRuntime.recoverParentICV(icv_previous__OMP_ParallelRegion_0);
        RuntimeException OMP_ee_0 = (RuntimeException) _OMP_ParallelRegion_0_in.OMP_CurrentParallelRegionExceptionSlot.get();
        if (OMP_ee_0 != null) {throw OMP_ee_0;}
        /*OpenMP Parallel region (#0) -- END */

        //#BEGIN GUI execution block
        if (SwingUtilities.isEventDispatchThread()) {
            repaint();}
        else {
            try {
    class _OMP_GuiCode_3 implements Runnable{
        _OMP_GuiCode_3(){
        }
        @Override
        public void run() {
            /****User Code BEGIN***/
            repaint();
            /****User Code END***/
        }
    }
    SwingUtilities.invokeLater(new _OMP_GuiCode_3());
} catch (Exception e) {e.printStackTrace();}

            }
            //#END GUI execution block

        }
        }
class _OMP_ParallelRegion_0{
        private int OMP_threadNumber = 1;
        private InternalControlVariables icv;
        private ReentrantLock OMP_lock;
        private Object OMP_reduction_lock = new Object();
        private ParIterator<?> OMP__ParIteratorCreator;
        public AtomicReference<Throwable> OMP_CurrentParallelRegionExceptionSlot = new AtomicReference<Throwable>(null);

        //#BEGIN shared, private, lastprivate, reduction variables defined here
        public double yTick;
        public int width;
        public int hight;
        public double y;
        public double xTick;
        public double x;
        //#END shared, private, lastprivate, reduction variables defined here
        public _OMP_ParallelRegion_0(int thread_num, InternalControlVariables icv) {
            this.icv = icv;
            if ((false == Pyjama.omp_get_nested()) && (Pyjama.omp_get_level() > 0)) {
                this.OMP_threadNumber = 1;
            }else {
                this.OMP_threadNumber = thread_num;
            }
            icv.currentParallelRegionThreadNumber = this.OMP_threadNumber;
            icv.OMP_CurrentParallelRegionBarrier = new PjCyclicBarrier(this.OMP_threadNumber);
        }

        class MyCallable implements Callable<Void> {
            private int alias_id;
            //#BEGIN private/firstprivate reduction variables defined and initialised here
            //#END private/firstprivate reduction variables defined and initialised here
            MyCallable(int id){
                this.alias_id = id;
            }

            @Override
            public Void call() {
                try {
                    /****User Code BEGIN***/
                    /*OpenMP Work Share region (#1) -- START */
                    
                {//#BEGIN firstprivate lastprivate reduction variables defined and initialized here
                    //#set implicit barrier here, otherwise unexpected initial value happens
                    PjRuntime.setBarrier();
                    //#END firstprivate lastprivate reduction variables defined and initialized here
                    try{
                        int i=0;
                        int OMP_iterator = 0;
                        int OMP_end = (int)((hight)-(0))/(1);
                        if (((hight)-(0))%(1) == 0) {
                            OMP_end = OMP_end - 1;
                        }
                        int OMP_local_iterator = 0;
                        int OMP_Chunk_Starting_point = 0;
                        int OMP_Default_chunkSize_autoGenerated = (OMP_end+1)/Pyjama.omp_get_num_threads();
                        if (Pyjama.omp_get_thread_num() < (OMP_end+1) % Pyjama.omp_get_num_threads()) {
                            ++OMP_Default_chunkSize_autoGenerated;
                            OMP_Chunk_Starting_point = Pyjama.omp_get_thread_num() * OMP_Default_chunkSize_autoGenerated;
                        } else {
                            OMP_Chunk_Starting_point = Pyjama.omp_get_thread_num() * OMP_Default_chunkSize_autoGenerated + (OMP_end+1) % Pyjama.omp_get_num_threads();
                        }
                        for (OMP_local_iterator=OMP_Chunk_Starting_point; OMP_local_iterator<OMP_Chunk_Starting_point+OMP_Default_chunkSize_autoGenerated && OMP_Default_chunkSize_autoGenerated>0; ++OMP_local_iterator) {
                            i = 0 + OMP_local_iterator * (1);
                            {
                                for (int j = 0; j < width; j++) {
                                    x = xmin + xTick * j;
                                    y = ymin + yTick * i;
                                    drawPixel(j, i, x, y);
                                    //#BEGIN GUI execution block
                                    if (SwingUtilities.isEventDispatchThread()) {
                                        repaint();}
                                    else {
                                        try {
    class _OMP_GuiCode_2 implements Runnable{
        _OMP_GuiCode_2(){
        }
        @Override
        public void run() {
            /****User Code BEGIN***/
            repaint();
            /****User Code END***/
        }
    }
    SwingUtilities.invokeLater(new _OMP_GuiCode_2());
} catch (Exception e) {e.printStackTrace();}

                                        }
                                        //#END GUI execution block

                                    }
                                }if (OMP_end == OMP_local_iterator) {
                                    //BEGIN lastprivate variables value set
                                    //END lastprivate variables value set
                                }
                            }
                        } catch (pj.pr.exceptions.OmpWorksharingLocalCancellationException e){
                            //TODO: work redistribution
                            Exception OMP_registered_e = e.getThrow();
                            if (null != OMP_registered_e) {
                                throw OMP_registered_e;
                            }
                        } catch (pj.pr.exceptions.OmpWorksharingGlobalCancellationException e){
                            //just escape from the work sharing region and wait other thread in next barrier
                            PjExecutor.cancelCurrentWorksharing();
                            Exception OMP_registered_e = e.getThrow();
                            if (null != OMP_registered_e) {
                                throw OMP_registered_e;
                            }
                        } catch (Exception e){throw e;}
                        //BEGIN  reduction
                        PjRuntime.reductionLockForWorksharing.lock();
                        PjRuntime.reductionLockForWorksharing.unlock();//END reduction
                        PjRuntime.setBarrier();
                    }

                    PjRuntime.setBarrier();
                    PjRuntime.reset_OMP_orderCursor();
                    /*OpenMP Work Share region (#1) -- END */

                    /****User Code END***/
                    //BEGIN reduction procedure
                    //END reduction procedure
                    //There is an implicit PjRuntime.setBarrier() inside PjRuntime.taskWait();
                    PjRuntime.taskWait();
                } catch (OmpParallelRegionLocalCancellationException e) {
                    PjRuntime.decreaseBarrierCount();
                    if (null != e.getThrow()) {throw e.getThrow();}
                } catch (OmpParallelRegionGlobalCancellationException e) {
                    PjRuntime.decreaseBarrierCount();
                    PjExecutor.cancelCurrentThreadGroup();
                    Exception OMP_registered_e = e.getThrow();
                    if (null != OMP_registered_e) {
                        OMP_CurrentParallelRegionExceptionSlot.compareAndSet(null, OMP_registered_e);
                    }
                } catch (Exception e) {
                    PjRuntime.decreaseBarrierCount();
                    PjExecutor.cancelCurrentThreadGroup();
                    OMP_CurrentParallelRegionExceptionSlot.compareAndSet(null, e);
                }
                if (0 == this.alias_id) {
                }
                return null;
            }
        }
        public void runParallelCode() {
            for (int i = 1; i <= this.OMP_threadNumber-1; i++) {
                Callable<Void> slaveThread = new MyCallable(i);
                PjRuntime.submit(i, slaveThread, icv);
            }
            Callable<Void> masterThread = new MyCallable(0);
            PjRuntime.getCurrentThreadICV().currentThreadAliasID = 0;
            try {
                masterThread.call();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }




        private void drawPixel(int i, int j, double x, double y) {{
            int colorIndex = MandelbrotCal(x, y);
            this.raster.setPixel(i, j, HSBtoRGB(colorIndex));
        }
        }


        private int[] HSBtoRGB(int colorIndex) {{
            int rgb = Color.HSBtoRGB(colorIndex / 100F, 1F, 1F);
            int[] RGBpixel = new int[3];
            RGBpixel[0] = (rgb >> 16) & 0xFF;
            RGBpixel[1] = (rgb >> 8) & 0xFF;
            RGBpixel[2] = rgb & 0xFF;
            return RGBpixel;
        }
        }


        private int MandelbrotCal(double x, double y) {{
            int count = 0;
            double zx = x;
            double zy = y;
            while (count < iterationDept && Math.abs(x) < 100 && Math.abs(zy) < 100) {
                double new_zx = zx * zx - zy * zy + x;
                zy = 2 * zx * zy + y;
                zx = new_zx;
                count++;
            }
            return count;
        }
        }


        public static void main(String[] s) {{
            new PjMandelbrotPar();
        }
        }


        @Override
        public void mouseClicked(MouseEvent e) {{
            long startTime = System.nanoTime();
            /*OpenMP Target region (#0) -- START */
            _OMP_TargetTaskRegion_0 _OMP_TargetTaskRegion_0_in = new _OMP_TargetTaskRegion_0();
            _OMP_TargetTaskRegion_0_in.startTime = startTime;
            _OMP_TargetTaskRegion_0_in.e = e;
            if (PjRuntime.currentThreadIsTheTarget("worker")) {
                _OMP_TargetTaskRegion_0_in.run();
                startTime = _OMP_TargetTaskRegion_0_in.startTime;
                e = _OMP_TargetTaskRegion_0_in.e;
            } else {
                PjRuntime.submitTargetTask(Thread.currentThread(), "worker", _OMP_TargetTaskRegion_0_in);
            }
            /*OpenMP Target region (#0) -- END */

        }
        }
class _OMP_TargetTaskRegion_0 extends pj.pr.task.TargetTask<Void>{

    //#BEGIN shared, private variables defined here
    public long startTime;
    public MouseEvent e;
    //#END shared, private variables defined here

    private int OMP_state = 0;
    @Override
    public Void call() {
        try {
            /****User Code BEGIN***/
            {
                mouseX = e.getX();
                mouseY = e.getY();
                if (SwingUtilities.isLeftMouseButton(e)) {
                    updatePosition(ZoomType.zoomIn);
                } else if (SwingUtilities.isRightMouseButton(e)) {
                    updatePosition(ZoomType.zoomOut);
                }
                double timeElapse = (System.nanoTime() - startTime) / 10e6;
                System.out.println("done within:" + timeElapse + "ms");
            }
            /****User Code END***/
        } catch(pj.pr.exceptions.OmpCancelCurrentTaskException e) {
            ;
        }
        this.setFinish();
        return null;
    }
    
}





        @Override
        public void mouseEntered(MouseEvent e) {{
        }
        }


        @Override
        public void mouseExited(MouseEvent e) {{
        }
        }


        @Override
        public void mousePressed(MouseEvent e) {{
        }
        }


        @Override
        public void mouseReleased(MouseEvent e) {{
        }
        }


        @Override
        public void mouseWheelMoved(MouseWheelEvent e) {{
            long startTime = System.nanoTime();
            /*OpenMP Target region (#1) -- START */
            _OMP_TargetTaskRegion_1 _OMP_TargetTaskRegion_1_in = new _OMP_TargetTaskRegion_1();
            _OMP_TargetTaskRegion_1_in.startTime = startTime;
            _OMP_TargetTaskRegion_1_in.e = e;
            if (PjRuntime.currentThreadIsTheTarget("worker")) {
                _OMP_TargetTaskRegion_1_in.run();
                startTime = _OMP_TargetTaskRegion_1_in.startTime;
                e = _OMP_TargetTaskRegion_1_in.e;
            } else {
                PjRuntime.submitTargetTask(Thread.currentThread(), "worker", _OMP_TargetTaskRegion_1_in);
            }
            /*OpenMP Target region (#1) -- END */

        }
        }
class _OMP_TargetTaskRegion_1 extends pj.pr.task.TargetTask<Void>{

    //#BEGIN shared, private variables defined here
    public long startTime;
    public MouseWheelEvent e;
    //#END shared, private variables defined here

    private int OMP_state = 0;
    @Override
    public Void call() {
        try {
            /****User Code BEGIN***/
            {
                mouseX = e.getX();
                mouseY = e.getY();
                int notches = e.getWheelRotation();
                if (notches < 0) {
                    System.out.println("Zoom in");
                    updatePosition(ZoomType.zoomIn);
                } else if (notches > 0) {
                    System.out.println("Zoom out");
                    updatePosition(ZoomType.zoomOut);
                }
                double timeElapse = (System.nanoTime() - startTime) / 10e6;
                System.out.println("done within:" + timeElapse + "ms");
            }
            /****User Code END***/
        } catch(pj.pr.exceptions.OmpCancelCurrentTaskException e) {
            ;
        }
        this.setFinish();
        return null;
    }
    
}





        @Override
        public void keyTyped(KeyEvent e) {{
        }
        }


        @Override
        public void keyReleased(KeyEvent e) {{
        }
        }


        @Override
        public void keyPressed(KeyEvent e) {{
            long startTime = System.nanoTime();
            int key = e.getKeyCode();
            if (key == KeyEvent.VK_SPACE) {
                /*OpenMP Target region (#2) -- START */
                _OMP_TargetTaskRegion_2 _OMP_TargetTaskRegion_2_in = new _OMP_TargetTaskRegion_2();
                _OMP_TargetTaskRegion_2_in.startTime = startTime;
                if (PjRuntime.currentThreadIsTheTarget("worker")) {
                    _OMP_TargetTaskRegion_2_in.run();
                    startTime = _OMP_TargetTaskRegion_2_in.startTime;
                } else {
                    PjRuntime.submitTargetTask(Thread.currentThread(), "worker", _OMP_TargetTaskRegion_2_in);
                }
                /*OpenMP Target region (#2) -- END */

            }
        }
        }
class _OMP_TargetTaskRegion_2 extends pj.pr.task.TargetTask<Void>{

    //#BEGIN shared, private variables defined here
    public long startTime;
    //#END shared, private variables defined here

    private int OMP_state = 0;
    @Override
    public Void call() {
        try {
            /****User Code BEGIN***/
            {
                xmin = MandelbrotDefault.xmin;
                xmax = MandelbrotDefault.xmax;
                ymin = MandelbrotDefault.ymin;
                ymax = MandelbrotDefault.ymax;
                mouseX = (int) frameWidth / 2;
                mouseY = (int) frameHight / 2;
                updatePosition(ZoomType.noZoom);
                double timeElapse = (System.nanoTime() - startTime) / 10e6;
                System.out.println("reset within:" + timeElapse + "ms");
            }
            /****User Code END***/
        } catch(pj.pr.exceptions.OmpCancelCurrentTaskException e) {
            ;
        }
        this.setFinish();
        return null;
    }
    
}




    }
